((c-mode .   ((indent-tabs-mode . t)
              (c-basic-offset . 8)
              (tab-width . 8)))
 (c++-mode . ((indent-tabs-mode . t)
              (c-basic-offset . 8)
              (tab-width . 8)))
)
